/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994,1995  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   avi.c
 *
 *  PURPOSE:
 *	  contains AVI-specific code for the Distributed Compression client
 *
 *  FUNCTIONS:
 *    InitStreams(void)
 *    InitVideoOutputStream(void)
 *    DisplayFileInfo(PAVIFILE pfile)
 *    DisplayStreamInfo(void)
 *    WriteBunchToStreams(LONG lVidStream, LONG lBunchId, LONG lBunchLen,
 * 											LPBYTE lpbVidData, LONG alIndex[])
 *    SendBunchToDecompressor(void)
 *    MoveDecompBunchToCompressor(void)
 *    ReturnBunchFromCompressor(LONG lBunchId)
 *    ProcessBunches(void)
 *
 *  COMMENTS:
 *    The client divides a source AVI file into 'bunches' which are sent to a
 *    number of server processes running on remote machines.  The servers
 *    recompress the video, and the recompressed bunches are returned to the
 *    client to be re-assembled in the destination file.
 *
 *    In this case a video bunch means a collection of frames which can all be
 *    recompressed together, i.e. a key frame and all the frames up to but not
 *    including the next key frame.  The bunch size is determined by the
 *    occurrence of key frames in the original compressed source material.
 *
 *    The current implementation assumes that bunches partitioned in this way
 *    are completely independent of each other.
 *
 *    This file contains the code that deals with the AVI file and transfers
 *    video data to and from the servers via RPC.
 *
 *    '-Com-' refers to input (compressed) data
 *    '-Unc-' refers to intermediate (uncompressed) data
 *    '-Rec-' refers to output (recompressed) data
 *
 *    Assumes there is only one video stream in the input file.
 *
 *    The client application can be built to use either UNICODE or ANSI APIs,
 *    according to the project settings.
 */

#include	"dcomp.h"
#include	"dcomp_c.h"


// Macros for this file

// Pointers to the BITMAPINFOHEADER portions of the video formats

#define		glpbihCom	(&glpbiCom->bmiHeader)
#define		glpbihUnc	(&glpbiUnc->bmiHeader)
#define		glpbihRec	(&glpbiRec->bmiHeader)


//
//  FUNCTION: InitStreams
//
//  PURPOSE: initialise all the streams in the input file and create
//    corresponding streams in the output file
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    number of streams opened
//
//  COMMENTS:
//    For non-video streams, the output stream info and format and info are
//    also written.  For the video stream, InitVideoOutputStream is called.
//

LONG		InitStreams(void)
{
	LONG			lResult;
	LONG			lStream;
	LONG			lFmtSize;
	LPVOID			lFmt;
    AVISTREAMINFO	strhdr;
    AVISTREAMINFOW	strhdrw;	// WCHAR version of strhdr


    // Walk through all the streams in the file
    for (lStream = gcpavi; lStream < MAXNUMSTREAMS; lStream++)
    {
		// Open input stream
        gapaviSrc[lStream] = NULL;
        if (AVIFileGetStream(gpaviInFile, &gapaviSrc[lStream],
        									0L, lStream - gcpavi) != AVIERR_OK)
		{
			break;
		}

    	if (gapaviSrc[lStream] == NULL)
		{
        	break;
		}

		// Read stream info
		lResult = AVIStreamInfo(gapaviSrc[lStream], &strhdr, sizeof(strhdr));
		if (lResult != 0)
		{
			ErrExit(
                TEXT("Error code %ld (0x%lx) reading info for stream %d\n"),
				lResult, lResult, lStream);
		}


		// Deal with video stream
		if (strhdr.fccType == streamtypeVIDEO)
		{
			if (glVidStream == 0)
			{
				glVidStream = lStream;
			}

			if ((lResult = InitVideoOutputStream()) == 0)
			{
				continue;
			}
			else
			{
				Eprintf(
					TEXT("Error %ld(0x%08.8lx)")
					TEXT(" creating video output stream\n"),
					lResult, lResult);
			    break;
			}
		}


	    // It's non-video - set initial frames to zero as we are not offsetting
	    // the streams in this sample app
		strhdr.dwInitialFrames = 0;

	    
	    // create an output stream

#ifndef UNICODE
		// First create a dummy AVISTREAMINFOW stream header just for
		// AVIFileCreateStream()

		memcpy(&strhdrw, &strhdr, sizeof strhdr);
		if (MultiByteToWideChar(CP_ACP, 0UL,
				strhdr.szName, strlen(strhdr.szName) + 1, strhdrw.szName, 64)
																		== 0)
		{
			Eprintf(TEXT("Couldn't create AVISTREAMINFOW structure")
					TEXT(" for AVIFileCreateStream()\n"));
			break;
		}

	    lResult = AVIFileCreateStream(gpaviOutFile, &gapaviDest[lStream],
	    															&strhdrw);
#else
	    lResult = AVIFileCreateStream(gpaviOutFile, &gapaviDest[lStream],
	    															&strhdr);
#endif
	    if (lResult != 0)	// Stream created OK?
	    {
			Eprintf(TEXT("Error %ld(0x%08.8lx) creating output stream %ld\n"),
													lResult, lResult, lStream);
		    break;
	    }
	

		// Determine size of stream format data and allocate space for it
		lResult = AVIStreamFormatSize(gapaviSrc[lStream], 0, &lFmtSize);
		if (lResult != 0L)
		{
			Eprintf(
				TEXT("Error code %ld (0x%lx) reading stream format size\n"),
															lResult, lResult);
			break;
		}

		if ((lFmt = midl_user_allocate(lFmtSize)) == NULL)
		{
			Eprintf(
				TEXT("Error allocating format buffer for input stream %ld\n"),
																	 lStream);
			break;
		}


		// Read the stream format
		lResult = AVIStreamReadFormat(gapaviSrc[lStream], 0, lFmt, &lFmtSize);
		if (lResult != 0L)
		{
			Eprintf(
				TEXT("Error code %ld (0x%lx) reading stream format\n"),
															lResult, lResult);
			break;
		}


	    // Set format of new stream
	    lResult = AVIStreamSetFormat(gapaviDest[lStream], 0, lFmt, lFmtSize);
	    if (lResult != 0)
	    { 
			Eprintf(TEXT("Error setting format for output stream %ld\n"),
																	lStream);
		    break;
	    }


		// Free the format buffer
		midl_user_free(lFmt);


		// Buffer size for all non-video streams will be biggest suggested
		// buffer size
	 	gcbBufSize = max(gcbBufSize, strhdr.dwSuggestedBufferSize);
    }

    if (gcpavi == lStream)
    {
        Eprintf(TEXT("No streams found in inout file\n"));
    }

	lStream = lStream - gcpavi;
	gcpavi += lStream;
	return(lStream);
}


//
//  FUNCTION: InitVideoOutputStream
//
//  PURPOSE: initialise the video output stream
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    non-zero for error
//
//  COMMENTS:
//    The input and output streams are specified by global variables.
//    InitVideoOutputStream initialises global format buffers.
//

static LONG		InitVideoOutputStream(void)
{
    HRESULT				hr;
    AVISTREAMINFO		strhdr;
    AVISTREAMINFOW		strhdrw;	// WCHAR version of strhdr
    LONG				lBiSize;


    // Determine the size of the format data and allocate space for structures
	// N.B. this is done once only for the stream

    if ((hr = AVIStreamFormatSize(gapaviSrc[glVidStream], 0, &glBiSize)) != 0L)
	{
		Eprintf(TEXT("Error code %ld (0x%lx) reading stream format size\n"),
																	hr, hr);
		return(hr);
	}


	if ((glpbiCom = midl_user_allocate(glBiSize)) == NULL)
	{
		Eprintf(TEXT("Error allocating input stream format buffer\n"));
		return(-1);
	}

    
	if ((glpbiUnc = midl_user_allocate(glBiSize)) == NULL)
	{
		Eprintf(TEXT("Error allocating intermediate stream format buffer\n"));
		return(-1);
	}

    
	if ((glpbiRec = midl_user_allocate(glBiSize)) == NULL)
	{
		Eprintf(TEXT("Error allocating output stream format buffer\n"));
		return(-1);
	}


	// Read the stream format of the source stream

	lBiSize = glBiSize;
    hr = AVIStreamReadFormat(gapaviSrc[glVidStream], 0, glpbiCom, &lBiSize);
    if (hr != 0L)
	{
		Eprintf(TEXT("Error code %ld (0x%lx) reading stream format\n"),
																	hr, hr);
		return(hr);
	}


    // Display it for debugging

 	Eprintf(TEXT("Size of format data is %ld\n"), glBiSize);
    Eprintf(	TEXT("Bitmap info header\n\n")
				TEXT("biSize:\t\t%lu\n")
				TEXT("biWidth:\t%ld\n")
				TEXT("biHeight:\t%ld\n")
				TEXT("biPlanes:\t%lu\n")
				TEXT("biBitCount:\t%lu\n")
				TEXT("biCompression:\t%lu (0x%08.8lx) %4.4hs\n")
				TEXT("biSizeImage:\t%lu\n")
				TEXT("biXPelsPerMeter:\t%ld\n")
				TEXT("biYPelsPerMeter:\t%ld\n")
				TEXT("biClrUsed:\t%lu\n")
				TEXT("biClrImportant:\t%lu\n\n"),
				glpbihCom->biSize,
				glpbihCom->biWidth,
				glpbihCom->biHeight,
				glpbihCom->biPlanes,
				glpbihCom->biBitCount,
				glpbihCom->biCompression, glpbihCom->biCompression,
										&glpbihCom->biCompression,
				glpbihCom->biSizeImage,
				glpbihCom->biXPelsPerMeter,
				glpbihCom->biYPelsPerMeter,
				glpbihCom->biClrUsed,
				glpbihCom->biClrImportant
		   );


	// Read the stream info of the source stream

    hr = AVIStreamInfo(gapaviSrc[glVidStream], &strhdr, sizeof(strhdr));
    if (hr != 0L)
 	{
		Eprintf(TEXT("Error code %ld (0x%lx) reading stream info\n"), hr, hr);
		return(hr);
	}


    // Display it for debugging

	Dprintf(TEXT("\nStream info\n\n")
			TEXT("fccType:\t%4.4hs\n")
			TEXT("fccHandler:\t%4.4hs\n")
			TEXT("dwFlags:\t0x%08.8lx\n")
			TEXT("dwCaps:\t\t0x%08.8lx\n")
			TEXT("wPriority:\t%u\n")
			TEXT("wLanguage:\t%u\n")
			TEXT("dwScale:\t%lu\n")
			TEXT("dwRate:\t\t%lu\n")
			TEXT("dwStart:\t%lu\n")
			TEXT("dwLength:\t%lu\n")
			TEXT("dwInitialFrames:\t%lu\n")
			TEXT("dwSuggestedBufferSize:\t%lu\n")
			TEXT("dwQuality:\t%lu\n")
			TEXT("dwSampleSize:\t%lu\n")
			TEXT("rcFrame:\tleft %ld, top %ld, right %ld, bottom %ld\n")
			TEXT("dwEditCount:\t%lu\n")
			TEXT("dwFormatChangeCount:\t%lu\n")
			TEXT("szName:\t\t\"%s\"\n\n"),
			&strhdr.fccType,
			&strhdr.fccHandler,
			strhdr.dwFlags,
			strhdr.dwCaps,
			strhdr.wPriority,
			strhdr.wLanguage,
			strhdr.dwScale,
			strhdr.dwRate,
			strhdr.dwStart,
			strhdr.dwLength,
			strhdr.dwInitialFrames,
			strhdr.dwSuggestedBufferSize,
			strhdr.dwQuality,
			strhdr.dwSampleSize,
			strhdr.rcFrame.left, strhdr.rcFrame.top, strhdr.rcFrame.right,
														strhdr.rcFrame.bottom,
			strhdr.dwEditCount,
			strhdr.dwFormatChangeCount,
			strhdr.szName
		   );


    // Set parameters for the uncompressed stream

	memcpy(glpbihUnc, glpbihCom, glBiSize);
	glpbihUnc->biCompression = BI_RGB;
	if ((gdwOutCompression == mmioFOURCC('M', 'S', 'V', 'C'))
			&& (glpbihCom->biBitCount > 16))
	{
		glpbihUnc->biBitCount = 16;
	}

    glpbihUnc->biSizeImage =
    		((((UINT)glpbihUnc->biBitCount * glpbihUnc->biWidth + 31)&~31) / 8)
    													* glpbihUnc->biHeight;


    // Set parameters for the recompressed output stream
	// Image size etc. same as for the compressed original for now

	memcpy(glpbihRec, glpbihCom, glBiSize);
	glpbihRec->biCompression = gdwOutCompression;
	if ((gdwOutCompression == mmioFOURCC('M', 'S', 'V', 'C'))
			&& (glpbihCom->biBitCount > 16))
	{
		glpbihRec->biBitCount = 16;
	}

	strhdr.fccHandler = gdwOutCompression;
	SetRect(&strhdr.rcFrame, 0, 0, (int) glpbihRec->biWidth, 
            (int) glpbihRec->biHeight);


    // Create a stream

#ifndef UNICODE
	// First create a dummy AVISTREAMINFOW stream header just for
	// AVIFileCreateStream()

	memcpy(&strhdrw, &strhdr, sizeof strhdr);
	if (MultiByteToWideChar(CP_ACP, 0UL,
			strhdr.szName, strlen(strhdr.szName) + 1, strhdrw.szName, 64) == 0)
	{
		Eprintf(TEXT("Couldn't create AVISTREAMINFOW structure")
				TEXT(" for AVIFileCreateStream()\n"));
		return(hr);
	}

    hr = AVIFileCreateStream(gpaviOutFile, &gapaviDest[glVidStream], &strhdrw);
#else
    hr = AVIFileCreateStream(gpaviOutFile, &gapaviDest[glVidStream], &strhdr);
#endif
    if (hr != 0)	// Stream created OK?
    {
	    return(hr);
    }


    // Set format of new stream

    hr = AVIStreamSetFormat(gapaviDest[glVidStream], 0, glpbiRec, lBiSize);
    if (hr != 0)
    { 
	    return(hr);
    }
}


//
//  FUNCTION: DisplayFileInfo
//
//  PURPOSE: display the AVI file header info for debugging
//
//  PARAMETERS:
//    pfile - pointer to open AVI file
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void		DisplayFileInfo(PAVIFILE pfile)
{
	AVIFILEINFO	fi;
	LONG 		lResult;

	if ((lResult = AVIFileInfo(pfile, &fi, sizeof(fi))) != 0)
	{
		Eprintf(TEXT("Error code %ld (0x%lx) reading AVI file info\n"),
                                                             lResult, lResult);
		return;
	}

	Dprintf(
        TEXT("File info\ndwFlags: 0x%lx\n")
        TEXT("dwStreams: %ld\nszFileType: \"%s\"\n"),
		fi.dwFlags, fi.dwStreams, fi.szFileType);
}


//
//  FUNCTION: DisplayStreamInfo
//
//  PURPOSE: display the AVI stream header info of all streams in the input
//    file, for debugging
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void		DisplayStreamInfo(void)
{
    AVISTREAMINFO	avis;
    LONG			lResult;
    int				i;

    // Walk through all streams loaded, reading and displaying stream info

    for (i = 0; i < gcpavi; i++)
    {
		if((lResult = AVIStreamInfo(gapaviSrc[i], &avis, sizeof(avis))) != 0)
		{
			Eprintf(
                TEXT("Error code %ld (0x%lx) reading info for stream %d\n"),
														lResult, lResult, i);
			return;
		}

		Eprintf(TEXT("\nStream %d\nType: %4.4hs\n"), i, &avis.fccType);
	}
}


//
//  FUNCTION: WriteBunchToStreams
//
//  PURPOSE: write a completed video bunch, with its associated non-video data
//           (mainly audio samples) to the output file
//
//  PARAMETERS:
//    lVidStream - stream number of the video stream
//    lBunchId   - bunch ID (equal to first video frame number in bunch)
//    lBunchLen  - number of video frames in bunch
//    lpbVidData - pointer to buffer containing the compressed video data
//    alIndex[]  - array of indexes of the video frames in the buffer
//
//  RETURN VALUE:
//    non-zero for error
//
//  COMMENTS:
//    
//

LONG	WriteBunchToStreams
(
	LONG	lVidStream,
	LONG	lBunchId,
	LONG	lBunchLen,
	LPBYTE	lpbVidData,
	LONG	alIndex[]
)
{
	LONG		lThisVideoFrame;
	LONG		lThisRecStartTime;
	LONG		lNextRecStartTime;
	LONG		lThisRecStartSample;
	LONG		lNextRecStartSample;
	LONG		lResult;
	LONG		l;
	LONG		lStream;
	LPBYTE		lpbBuffer = NULL;
	LONG		lBytes;
	LONG		lNumSamples;


	// Allocate a buffer based on suggested buffer size from all the streams

	if ((lpbBuffer = midl_user_allocate(gcbBufSize)) == NULL)
	{
		Eprintf(TEXT("Couldn't allocate memory for non-video samples"));
		return(-1);
	}


	// Walk through the frames in the bunch

	for (l = 0; l < lBunchLen; l++)
	{
		// Calculate the time-equivalent of this stream position

		lThisVideoFrame = lBunchId + l;
		lThisRecStartTime = AVIStreamSampleToTime(gapaviSrc[lVidStream],
														lThisVideoFrame);
		lNextRecStartTime = AVIStreamSampleToTime(gapaviSrc[lVidStream],
														lThisVideoFrame + 1);


		// Write the video frame to the output file

		Dprintf(	TEXT("AVIStreamWrite(0x%08.8lx, %ld, %ld, 0x%08.8lx,")
		            TEXT(" %ld, 0x%08.8lx, 0x%08.8lx, 0x%08.8lx)\n"),
		    		gapaviDest[lVidStream],
		    		lThisVideoFrame,
		    		1, &lpbVidData[alIndex[l]],
		    		alIndex[l + 1] - alIndex[l],
		    		(l == 0 ? AVIIF_KEYFRAME : 0UL),
		    		NULL,
		    		NULL);

		lResult = AVIStreamWrite
	    		  (
					gapaviDest[lVidStream],
					lThisVideoFrame,
					1,
					&lpbVidData[alIndex[l]],
					alIndex[l + 1] - alIndex[l],
					(l == 0 ? AVIIF_KEYFRAME : 0UL),
					NULL,
					NULL
				  );
	    if (lResult  != 0L)
	 	{
			Eprintf(
				TEXT("Error code %ld(0x%lx)")
                TEXT(" writing frame %ld to stream %ld\n"),
				lResult, lResult, lThisVideoFrame, lVidStream);
			goto error;
		}


		// Walk through the non-video frames, copying the samples for the
		// equivalent time interval from the input to the output file

		for (lStream = 0; lStream < gcpavi; lStream++)
		{
			if (lStream == lVidStream)
			{
				continue;
			}


			// Calculate the start and finish sample numbers corresponding to
			// this video frame period

			lThisRecStartSample = min
								  (
								  	AVIStreamTimeToSample(gapaviSrc[lStream],
								  							lThisRecStartTime),
									AVIStreamEnd(gapaviSrc[lStream])
								  );

			lNextRecStartSample = min
								  (
								  	AVIStreamTimeToSample(gapaviSrc[lStream],
								  							lNextRecStartTime),
									AVIStreamEnd(gapaviSrc[lStream])
								  );


			// Copy the samples (if any)

			if ((lNumSamples = lNextRecStartSample - lThisRecStartSample) > 0)
			{
				// Read a block of samples

				lResult = AVIStreamRead
						  (
							gapaviSrc[lStream],
							lThisRecStartSample,
							lNumSamples,
							lpbBuffer,
							gcbBufSize,
							&lBytes,
							NULL
						  );
				if (lResult != 0L)
				{
					Eprintf(
						TEXT("Error code %ld(0x%08.8lx)")
						TEXT(" reading sample %ld of stream %ld\n"),
						lResult, lResult, lThisRecStartSample, lStream);
					goto error;
				}

		 		Dprintf(TEXT("Read %ld bytes\n"), lBytes);


				// Write it

				lResult = AVIStreamWrite
						  (
							gapaviDest[lStream],
							lThisRecStartSample,
							lNumSamples,
							lpbBuffer,
							lBytes,
							0UL,
							NULL,
							NULL
						  );
				if (lResult != 0L)
				{
					Eprintf(
						TEXT("Error code %ld(0x%lx)")
						TEXT(" writing frame %ld of stream %ld\n"),
						lResult, lResult, lThisRecStartSample, lStream);
					goto error;
				}
				else
				{
			 		Dprintf(TEXT("Wrote %ld bytes\n"), lBytes);
				}
			}
		}


		// Mark the end of the chunk for this video frame

		if ((lResult = AVIFileEndRecord(gpaviOutFile)) != 0)
		{
			Eprintf(
				TEXT("Error code %ld(0x%lx)")
				TEXT(" writing end-of-record\n"),
				lResult, lResult);
			goto error;
		}
	}

	midl_user_free(lpbBuffer);
	return(0);


error:
	midl_user_free(lpbBuffer);
	return(-1);
}


//
//  FUNCTION: SendBunchToDecompressor
//
//  PURPOSE:
//    send a bunch to the server for decompression
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    Number of frames in sent bunch (0 if failed to send bunch)
//    Negative on fatal error
//
//  COMMENTS:
//    If gbaAction is RECOMPRESS, the bunch will be compressed at the same
//    server after decompression.  If it is DECOMPRESS, the uncompressed bunch
//    will be queued for return to the client.
//

LONG		SendBunchToDecompressor(LONG lBunchId)
{
	static LONG				fBunchReady = FALSE;
    static LONG				lThisKey;
	static LPBYTE			lpbComData;
    static LONG				alIndex[MAXBUNCHLEN + 1];

    LONG					lNextKey;
	LONG					lBunchLen;
    LONG					l;
	LONG					lBytes;
    HRESULT					hr;
	RPC_STATUS RPC_ENTRY	status;


	// Read the bunch from file if we don't already have it

	if
	(
			!fBunchReady
		||
			(fBunchReady && (lBunchId != lThisKey))
	)
	{
		// Find size of bunch

		lThisKey = lBunchId;
		lNextKey = AVIStreamNextKeyFrame(gapaviSrc[glVidStream], lThisKey);

		if (lNextKey < 0)
		{
			// If no more key frames, set NExtKey to the end of the stream
			lNextKey = AVIStreamEnd(gapaviSrc[glVidStream]);
		}

		lBunchLen = lNextKey - lThisKey;
		if (lBunchLen > MAXBUNCHLEN)
		{
			Eprintf(
	            TEXT("Bunch too big: beginning frame %ld,")
	            TEXT(" length %ld\n"),
				lThisKey, lBunchLen);
		    return(-1);
		}


	    // Allocate memory for the bitmaps

		Dprintf(
	        TEXT("ACT1: Allocating compressed buffer for bunch %ld\n"),
	        lBunchId);

		if ((lpbComData
	           = midl_user_allocate(
	// temp.: make this the full size of the uncompressed data for safety.
	// Should really get a suggested buffer size from compressor
									glpbihUnc->biSizeImage * lBunchLen
								   )) == NULL)
		{
			Eprintf(
				TEXT("Error allocating input buffer for bunch %ld\n"),
															lBunchId);
		    return(0);
		}


		// Fill input buffers from file

		alIndex[0] = 0L;
		for (l = 0; l < lBunchLen; l++)
		{
			Dprintf(TEXT("AVIStreamRead(0x%08.8lx, %ld, %ld, ")
					TEXT("0x%08.8lx, %ld, 0x%08.8lx, 0x%08.8lx)\n"),
		    		gapaviSrc[glVidStream],
		    		lBunchId + l,
		    		1,
		    		lpbComData + alIndex[l],
		    		glpbihCom->biSizeImage,
		    		&lBytes,
		    		NULL);

		    hr = AVIStreamRead
				(
					gapaviSrc[glVidStream],
					lBunchId + l,
					1,
					lpbComData + alIndex[l],
					glpbihCom->biSizeImage,
					&lBytes,
					NULL
				);

		    if (hr != 0L)
			{
				Dprintf(
	                TEXT("Error code %ld(0x%08.8lx)")
	                TEXT(" reading frame %ld\n"),
					hr, hr, lBunchId + l);
				return(-1);
			}

			Dprintf(TEXT("Read %ld bytes\n"), lBytes);

			alIndex[l + 1] = alIndex[l] + lBytes;
				// alIndex[lBunchLen] ends up with total size of data
		}

		fBunchReady = TRUE;
	}


	// Send the bunch out to be decompressed

	Dprintf(TEXT("Queueing bunch %ld for decompression\n"), lBunchId);

	RpcTryExcept
	{
		if ((hr = RPCDecompBunchRequest
				  (
					gahDecomp[lNextDecompSrv],
					lBunchId,
					gbaAction,
					lBunchLen,
					glBiSize,
					(BYTE*) glpbihCom,
					glBiSize,
					(BYTE*) glpbihUnc,
					glBiSize,
					(BYTE*) glpbihRec,
					alIndex,
					alIndex[lBunchLen],
					lpbComData
				  )) == 0)
		{
			// Success - prepare for next bunch

			Eprintf(
				TEXT("Queued bunch %ld for decompression on \"%s\"\n"),
	            lBunchId, gaszDecompServer[lNextDecompSrv]);
			fBunchReady = FALSE;	// clear away the bunch data
		}
		else
		{
			// Failure - keep the bunch hanging around for the next time

			Dprintf(
	            TEXT("Problem queueing bunch %ld for decompression\n"),
															 lBunchId);
			lBunchLen = 0;
		}
	}
	RpcExcept(1)
	{
		// Exception

		status = RpcExceptionCode();
	    Eprintf
	    (
	    	TEXT("Exception %ld (0x%08.8lx) queueing bunch %ld")
	        TEXT(" for decompression on \"%s\"\n"),
	    	status, status, lBunchId, gaszDecompServer[lNextDecompSrv]
	    );

		lBunchLen = 0;
	}
	RpcEndExcept;


	// Free memory for compressed bunch

	if (!fBunchReady)
	{
		Dprintf(
	        TEXT("ACT1: Freeing compressed buffer")
	        TEXT(" for bunch %ld at 0x%08.8lx\n"),
			lBunchId, lpbComData);
		midl_user_free(lpbComData);
	}

	return(lBunchLen);
}


//
//  FUNCTION: MoveDecompBunchToCompressor
//
//  PURPOSE:
//    retrieve an uncompressed bunch from a server and send it out to be
//    compressed on another server
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    Negative on fatal error, otherwise zero
//
//  COMMENTS:
//    This function is only needed when gbaAction is DECOMPRESS, in other
//    words, there are separate decompression and compression servers.  This
//    configuration is not all that useful since the cost of decompression is
//    relatively very small, and the extra network load is not worth it.
//
//    If the output from any decompressor can feed any compressor, a situation
//    can arise when a fast decompressor succeeds in clogging up all the
//    compression servers, and a slow decompressor can't pass on its finished
//    bunches, deadlocking the whole system.  This sample code adopts the very
//    simple (but not very efficient) strategy of pairing decompression and
//    compression servers to avoid this problem.  Other solutions are possible.
//

LONG		MoveDecompBunchToCompressor(void)
{
	LONG					lBunchId;
	LONG					lBunchLen;
	LONG					lUncDataLen;
	LPBYTE					lpbUncData;
	LONG					lReturnStatus;
    HRESULT					hr;
	RPC_STATUS RPC_ENTRY	status;


	// Allocate the uncompressed data buffer

	lUncDataLen = glpbihUnc->biSizeImage * MAXBUNCHLEN;
	                      // since we don't yet know the bunch length

	Dprintf(TEXT("ACT1: Allocating uncompressed buffer\n"));

	if ((lpbUncData = midl_user_allocate(lUncDataLen)) == NULL)
	{
		Eprintf(TEXT("Error allocating uncompressed buffer\n"));
	    return(0);
	}

	RpcTryExcept
	{
		lBunchId = -1;
		if ((hr = RPCGetDecompBunchResult
				  (
					gahDecomp[lNextDecompSrv],
					&lReturnStatus,
					&lBunchId,
					&lBunchLen,
					glBiSize,
					(BYTE*) glpbihUnc,
					glpbihUnc->biSizeImage * MAXBUNCHLEN,
					&lUncDataLen,
					lpbUncData
				  )) == 0)
		{
			// Success

			Eprintf(TEXT("Retrieved bunch %ld from decompressor\n"),
	                                                      lBunchId);
		}
		else
		{
			// Failure

			Eprintf(
	            TEXT("Problem retrieving bunch from decompressor\n"));
			hr = 1L;
		}
	}
	RpcExcept(1)
	{
		status = RpcExceptionCode();
	    Eprintf
	    (
	    	TEXT("Exception %ld (0x%08.8lx)")
	        TEXT(" retrieving bunch from decompressor\n"),
	    	status, status
	    );
		hr = 1L;
	}
	RpcEndExcept;


	// If we got a return bunch, queue it for recompression

	if (!hr)
	{
	    RpcTryExcept
	    {
			Dprintf(
	            TEXT("Queueing bunch %ld")
	            TEXT(" located at 0x%08.8lx for compression\n"),
							                  lBunchId, lpbUncData);
			if ((hr = RPCCompBunchRequest
					  (
					  	gahComp[lNextDecompSrv],
							// N.B. queue it on the same server number
							// as the decompressor, to avoid logjams
					  	lBunchId,
					  	lBunchLen,
						glBiSize,
						(BYTE*) glpbihUnc,
						glBiSize,
						(BYTE*) glpbihRec,
					  	lUncDataLen,
					  	lpbUncData
					  )) == 0)
			{
				// Free memory for compressed bunch.

				Eprintf(
					TEXT("Queued bunch %ld for compression")
					TEXT(" on \"%s\"\n"),
	                lBunchId, gaszCompServer[lNextDecompSrv]);

				Dprintf(
	                TEXT("Freeing uncompressed buffer")
	                TEXT("for bunch %ld at 0x%08.8lx\n"),
					lBunchId, lpbUncData);
				midl_user_free(lpbUncData);
			}
			else
			{
				// To do! Must re-try this request if it fails.
				// Storage for the uncompressed bunch should be static

				Eprintf(
					TEXT("Problem queueing bunch %ld")
	                TEXT(" for compression on \"%s\"\n"),
					lBunchId, gaszCompServer[lNextDecompSrv]);
			}
	    }
	    RpcExcept(1)
	    {
			status = RpcExceptionCode();
	        Eprintf
	        (
	        	TEXT("Exception %ld (0x%08.8lx) queueing")
	            TEXT(" bunch %ld for compression on \"%s\"\n"),
	        	status, status, lBunchId,
	        	gaszCompServer[lNextDecompSrv]
	        );
			hr = 1L;
	    }
	    RpcEndExcept;
	}

	return(0);
}


//
//  FUNCTION: ReturnBunchFromCompressor
//
//  PURPOSE:
//    retrieve a finished bunch from a server and write it to the destination
//    AVI file
//
//  PARAMETERS:
//    BunchId - ID of bunch to be returned
//
//  RETURN VALUE:
//    Number of frames in returned bunch (0 if failed to return bunch)
//    Negative on fatal error
//
//  COMMENTS:
//
//

LONG		ReturnBunchFromCompressor(LONG lBunchId)
{
	LONG					lBunchLen;
	LPBYTE					lpbRecData;
    LONG					lRecDataLen;
    LONG					alIndex[MAXBUNCHLEN + 1];
	DWORD					dwRetFlags[MAXBUNCHLEN];
	LONG					lReturnStatus;
    HRESULT					hr;
	RPC_STATUS RPC_ENTRY	status;


	if ((lpbRecData
	         = midl_user_allocate(
	// temp.: make this the full size of the uncompressed data for safety
								  glpbihUnc->biSizeImage * MAXBUNCHLEN
	                              )) == NULL)
	{
		Eprintf(
	        TEXT("Error allocating recompressed data buffer")
	        TEXT(" for bunch %ld\n"),
			lBunchId);
	    return(0);
	}

	Dprintf(TEXT("Retrieving recompressed bunch %ld\n"), lBunchId);

	RpcTryExcept
	{
		if ((hr = RPCGetCompBunchResult
				  (
					gahComp[lNextCompSrv],
					&lReturnStatus,
					&lBunchId,
					&lBunchLen,
					glBiSize,
					(BYTE*) glpbihRec,
					MAXBUNCHLEN,
					alIndex,
					dwRetFlags,
					glpbihUnc->biSizeImage * MAXBUNCHLEN,
							// Better: keep track of the longest bunch
							// length sent out so far, and use the
	                        // suggested buffer size of the compressor
					&lRecDataLen,
					lpbRecData
				  )) == 0)
		{
			Eprintf(
				TEXT("Retrieved bunch %ld")
				TEXT(" from compressor on \"%s\"\n"),
	            lBunchId, gaszCompServer[lNextCompSrv]);
		}
		else
		{
			Eprintf(
	            TEXT("Problem retrieving bunch %ld")
	            TEXT(" from compressor on \"%s\"\n"),
				lBunchId, gaszCompServer[lNextCompSrv]);
			lBunchLen = 0;
		}
	}
	RpcExcept(1)
	{
		status = RpcExceptionCode();
	    Eprintf
	    (
	    	TEXT("Exception %ld (0x%08.8lx)")
	        TEXT(" retrieving bunch %ld from compressor on \"%s\"\n"),
	    	status, status, lBunchId, gaszCompServer[lNextCompSrv]
	    );
		hr = -1;
		lBunchLen = 0;
	}
	RpcEndExcept;

	if (lReturnStatus != 0)
	{
	    Eprintf
	    (
	    	TEXT("Return status of recompressed bunch from \"%s\"")
	        TEXT(" was %ld (0x%08.8lx)\n"),
	        gaszCompServer[lNextCompSrv],
	    	lReturnStatus, lReturnStatus
	    );
		return(-1);
	}


	// Save the recompressed data (if any) and free the bitmap buffer

	if (!hr)
	{
		alIndex[lBunchLen] = lRecDataLen;

	    if ((hr = WriteBunchToStreams
				  (
					glVidStream,
					lBunchId,
					lBunchLen,
					lpbRecData,
					alIndex
				  )) != 0L)
	 	{
			Eprintf(
				TEXT("Error code %ld(0x%lx)")
	            TEXT(" writing bunch %ld to stream\n"),
				hr, hr, lBunchId);
			return(-1);
		}
	}

	Dprintf(
	    TEXT("Freeing compressed buffer for bunch %ld at 0x%08.8lx\n"),
		lBunchId, lpbRecData);

	midl_user_free(lpbRecData);
	return(lBunchLen);
}


//
//  FUNCTION: ProcessBunches
//
//  PURPOSE: recompress a video stream by sending bunches to servers, then
//    reassemble returned bunches into output stream 
//
//  PARAMETERS:
//    ps - pointer to open input stream
//    pf - pointer to open output AVI file
//
//  RETURN VALUE:
//
//  COMMENTS:
//    Create bunches of frames from the input compressed stream, based on the
//    incidence of keyframes in the stream, and first decompress, then compress
//    whole bunches by using per-bunch RPC calls to the servers.
//    If gbaAction is set to RECOMPRESS (the default), the servers decompress
//    and compress without returning the bunches to the client in between.
//
//    Currently the palette size is fixed for each stream and doesn't take
//    account of format changes.  The size of the BITMAPINFOHEADER is assumed
//    to be fixed - this is not valid for certain compression formats.
//
//    Currently the parameters for the uncompressed and recompressed streams
//    are based on the input format - only the output compression type is
//    selected by the user.  (However, when converting to MSVC, the
//    bits-per-pixel value is restricted to 16, because of the compressor
//    capabilities.  This is a quick fix: it should really be done by
//    negotiating with the compressor.)
//

void		ProcessBunches(void)
{
	LONG					lStreamEnd;
	LONG					lCompStatus;
	LONG					lDecompStatus;
	LONG					lReadBunch;
	LONG					lWriteBunch;
	LONG					lResult;
	RPC_STATUS RPC_ENTRY	status;


    lReadBunch = lWriteBunch = AVIStreamStart(gapaviSrc[glVidStream]);
	Dprintf(TEXT("First frame: %ld\n"), lReadBunch);
 	lStreamEnd = AVIStreamEnd(gapaviSrc[glVidStream]);
	lNextDecompSrv = lNextCompSrv = 0L;


	// Main loop transferring data between the various processing queues.
	// This loop does (up to) three transfer actions to each bunch:
	//	1. From the input AVI file to the decompressor input queue
	//	2: From the decompressor output queue to the compressor input queue
    //     (Not required in RECOMPRESS mode - the default)
	//	3: From the compressor output queue to the output AVI file
	// These actions are treated as asynchronous, and each is triggered 
	// by the status of the queues and files during each iteration of the loop.
	// These actions could instead be carried out in separate threads.

	for(;;)
    {
		// We want to poll each queue, say, four times per second - not too
		// often as this polling generates traffic on the net and context
		// switches on servers

		Sleep(250/lDecompSrvCnt);	// (normally lCompSrvCnt == lDecompSrvCnt)


		// Move on to next decompression server

		lNextDecompSrv = (lNextDecompSrv + 1) % lDecompSrvCnt;


		// Action 1

		// Queue an input bunch for decompression if there are any and the
		// decompressor queue has room. Do this until the input video data is
		// finished.

		if (lReadBunch < lStreamEnd)
		{
		    RpcTryExcept
		    {
				// Ask the server if it will accept a bunch

				lDecompStatus
						= RPCDecompBunchReqAccept(gahDecomp[lNextDecompSrv]);
		    }
		    RpcExcept(1)
		    {
				status = RpcExceptionCode();
		        Dprintf
		        (
		        	TEXT("Exception %ld (0x%08.8lx) getting queue status")
		        	TEXT(" from \"%s\" for Action 1\n"),
		        	status, status, gaszDecompServer[lNextDecompSrv]
		        );
				lDecompStatus = FALSE;
		    }
		    RpcEndExcept;

			if (lDecompStatus)
			{
				lResult = SendBunchToDecompressor(lReadBunch);
				if (lResult >= 0)
				{
					lReadBunch += lResult;
				}
				else
				{
					// Fatal error

					break;
				}
			}
		}


		// Action 2

		// Queue a decompressed bunch to be recompressed, if there are any and
		// if the compressor queue has room.  Don't need to do this in
		// RECOMPRESS mode.

		if (gbaAction != RECOMPRESS)
		{
		    RpcTryExcept
		    {
				// Check if decompression server has a bunch available and
				// compression server will accept a bunch

				lDecompStatus
                        = RPCDecompBunchRetAvail(gahDecomp[lNextDecompSrv]);
				lCompStatus
						= RPCCompBunchReqAccept(gahComp[lNextDecompSrv]);
		    }
		    RpcExcept(1)
		    {
				status = RpcExceptionCode();
		        Dprintf
		        (
		        	TEXT("Exception %ld (0x%08.8lx) getting queue status")
		        	TEXT(" values from \"%s\" or \"%s\" for Action 2\n"),
		        	status, status,
		        	gaszDecompServer[lNextDecompSrv],
		        	gaszCompServer[lNextDecompSrv]
		        );
				lDecompStatus = FALSE;
				lCompStatus   = FALSE;
		    }
		    RpcEndExcept;
		}
		else
		{
			lDecompStatus = FALSE;
			lCompStatus   = FALSE;
		}
	
		if (lDecompStatus && lCompStatus)
		{
			lResult = MoveDecompBunchToCompressor();
			if (lResult < 0)
			{
				// Fatal error

				break;
			}
		}


		// Action 3

		// Save next recompressed bunch in the output file, if it is ready

		// Move on to next compression server

		lNextCompSrv = (lNextCompSrv + 1) % lCompSrvCnt;

 	    RpcTryExcept
	    {
			// Ask server if it has the bunch we want

			lCompStatus
                = RPCCompBunchRetAvail(gahComp[lNextCompSrv], lWriteBunch);
	    }
	    RpcExcept(1)
	    {
			status = RpcExceptionCode();
	        Dprintf
	        (
	        	TEXT("Exception %ld (0x%08.8lx) getting queue status values")
	        	TEXT(" from \"%s\" for Action 3\n"),
	        	status, status, gaszCompServer[lNextCompSrv]
	        );
			lCompStatus = FALSE;
	    }
	    RpcEndExcept;

		if (lCompStatus)
		{
			lResult = ReturnBunchFromCompressor(lWriteBunch);
			if (lResult >= 0)
			{
				lWriteBunch += lResult;
			}
			else
			{
				// Fatal error

				break;
			}
		}

		if (lWriteBunch >= lStreamEnd)
		{
			// Finished the stream

			break;
		}
    }


 	// Free the format buffers

	midl_user_free(glpbiCom);
	midl_user_free(glpbiUnc);
	midl_user_free(glpbiRec);
}
