/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   comp.c
 *
 *  PURPOSE:
 *	  contains free-standing video bunch compression code
 *
 *  FUNCTIONS:
 *	  CompBunch(<Bunch data>)
 *
 *  COMMENTS:
 *    In this case a video bunch means a collection of frames which are all to
 *    be compressed together, i.e. a key frame and all the frames up to but not
 *    including the next key frame.
 *
 *    This is part of the distributed recompression tool, so typically
 *    compression takes place after decompression.  This means the bunch size
 *    is usually already determined, by the occurrence of key frames in the
 *    original compressed source material.
 *
 *    '-In-' refers to input (uncompressed) data
 *    '-Out-' refers to output (compressed) data
 */

#include	"dcomp.h"
#include	"dcomp_s.h"

HIC FindVideoCompressor
(
	LPBITMAPINFOHEADER lpbiIn,
	LPBITMAPINFOHEADER lpbiOut,
	WORD wFlags
);


//
//  FUNCTION: CompBunch
//
//  PURPOSE:
//    compress the frames in a bunch using ICM calls
//
//  PARAMETERS:
//    lBunchId      - unique bunch ID - frame number of first frame in bunch
//    lBunchLen     - no. of frames in the bunch
//    lpbiIn        - pointer to video format for input data
//    lpbiOut       - pointer to video format for output data
//    lpbInData[]   - pointer to the input video data buffer
//    alOut[]       - pointer to array to receive indexes of frame starts in
//                    output video buffer
//    dwRetFlags[]  - pointer to array to receive flag for each frame returned
//                    by the compressor
//    lplOutDataLen - pointer to variable to receive the total length of the
//                    output video data
//    lpbOutData[]  - pointer to buffer to receive the output video data
//
//  RETURN VALUE:
//    Zero on success, non-zero on error
//
//  COMMENTS:
//

extern HWND hWnd;

LONG		CompBunch
(
	LONG			lBunchId,
	LONG			lBunchLen,
	LPBITMAPINFO	lpbiIn,
	LPBITMAPINFO	lpbiOut,
	BYTE			lpbInData[],
	LONG			alOut[],
	DWORD			dwRetFlags[],
	LONG			*lplOutDataLen,
	BYTE			lpbOutData[]
)
{
	HIC					hIC;
	LONG				l;
    HRESULT				hr;
	DWORD				dwFlags;
	LPBITMAPINFOHEADER	lpbiPrev;
	LPVOID				lpPrev;
	DWORD				dwCkid;
	LPBITMAPINFOHEADER	lpbihIn = &lpbiIn->bmiHeader;
	LPBITMAPINFOHEADER	lpbihOut = &lpbiOut->bmiHeader;
	ICINFO				icinfo;
	DWORD				dwSaveSizeImage = lpbihIn->biSizeImage;

	// Open compressor

//	if ((hIC = ICLocate(ICTYPE_VIDEO, 0, // lpbihOut->biCompression,
	if ((hIC = FindVideoCompressor(
								lpbihIn, lpbihOut, ICMODE_COMPRESS)) != NULL)
	{
		// Initialise compressor

		if ((hr = ICCompressQuery(hIC, lpbihIn, lpbihOut)) != ICERR_OK)
		{
			Eprintf(
				TEXT("Couldn't get compressor format: error code %ld\n"), hr);
			if ((hr = ICClose(hIC)) != ICERR_OK)
			{
				Eprintf(TEXT("Error code %ld closing")
						 TEXT("compressor \"%4.4hs\"\n"),
						 hr, &lpbihOut->biCompression);
			}
			return(1L);
		}

 		ICGetInfo(hIC, &icinfo, sizeof icinfo);
		Dprintf(TEXT("Compressor flags: 0x%08.8lx\n"), icinfo.dwFlags);
		Eprintf(TEXT("%s: Compressing bunch %ld\n"),
									_tgetenv(TEXT("COMPUTERNAME")), lBunchId);

		if (ICCompressBegin(hIC, lpbiIn, lpbiOut) == ICERR_OK)
		{
			alOut[0] = 0L;


			// Walk through the frames in the bunch

			for (l = 0L; l < lBunchLen; l++)
			{
			    // Compress the data

				dwFlags  = (l == 0 ? ICCOMPRESS_KEYFRAME : 0UL);
								// N.B. assumes only one key frame in bunch
				lpbiPrev = (l == 0 || (icinfo.dwFlags & VIDCF_FASTTEMPORALC) ?
						NULL : lpbihIn);
				lpPrev   = (l == 0 || (icinfo.dwFlags & VIDCF_FASTTEMPORALC) ?
						NULL : &lpbInData[lpbihIn->biSizeImage * (l - 1)]);

		        if (ICCompress(	hIC,
		        				dwFlags,
		        				lpbihOut,
		        				&lpbOutData[alOut[l]],
		        				lpbihIn,
		        				&lpbInData[lpbihIn->biSizeImage * l],
								&dwCkid,
								&dwRetFlags[l],	// Note - these returned values
												// sometimes seem to be invalid
								lBunchId + l,
								0UL,
								(DWORD) -1,
								lpbiPrev,
								lpPrev
			        		  ) != ICERR_OK)
			    {
					Eprintf(
						TEXT("Problem compressing frame %ld of bunch\n"), l);
					if ((hr = ICClose(hIC)) != ICERR_OK)
					{
						Eprintf(
							TEXT("Error code %ld closing compressor ")
							TEXT("\"%4.4hs\"\n"),
										hr, &lpbihOut->biCompression);
					}
					return(1L);
		        }

				if (l == (lBunchLen - 1))
				{
					*lplOutDataLen = alOut[l] + lpbihOut->biSizeImage;
				}
				else
				{
					alOut[l + 1] = alOut[l] + lpbihOut->biSizeImage;
				}
			}


			// Finish compression session and close compressor
			 
		    if ((hr = ICCompressEnd(hIC)) != ICERR_OK)
			{
				Eprintf(TEXT("Error code %ld (0x%lx) ending compression\n"),
																	hr, hr);
				if ((hr = ICClose(hIC)) != ICERR_OK)
				{
					Eprintf(
						TEXT("Error code %ld closing compressor \"%4.4hs\"\n"),
									hr, &lpbihOut->biCompression);
				}
				return(1L);
			}
		}
		else
		{
			Eprintf(TEXT("Couldn't initialise compressor \"%4.4hs\"\n"),
										&lpbihOut->biCompression);
			if ((hr = ICClose(hIC)) != ICERR_OK)
			{
				Eprintf(TEXT("Error code %ld closing compressor \"%4.4hs\"\n"),
									hr, &lpbihOut->biCompression);
			}
			return(1L);
		}
	}
	else
	{
		Eprintf(TEXT("Couldn't open compressor \"%4.4hs\"\n"),
										&lpbihOut->biCompression);

	    Dprintf(	TEXT("Input bitmap info header\n\n")
					TEXT("biSize:\t\t%lu\n")
					TEXT("biWidth:\t%ld\n")
					TEXT("biHeight:\t%ld\n")
					TEXT("biPlanes:\t%lu\n")
					TEXT("biBitCount:\t%lu\n")
					TEXT("biCompression:\t%lu (0x%08.8lx) %4.4hs\n")
					TEXT("biSizeImage:\t%lu\n")
					TEXT("biXPelsPerMeter:\t%ld\n")
					TEXT("biYPelsPerMeter:\t%ld\n")
					TEXT("biClrUsed:\t%lu\n")
					TEXT("biClrImportant:\t%lu\n\n"),
					lpbihIn->biSize,
					lpbihIn->biWidth,
					lpbihIn->biHeight,
					lpbihIn->biPlanes,
					lpbihIn->biBitCount,
					lpbihIn->biCompression, lpbihIn->biCompression,
											&lpbihIn->biCompression,
					lpbihIn->biSizeImage,
					lpbihIn->biXPelsPerMeter,
					lpbihIn->biYPelsPerMeter,
					lpbihIn->biClrUsed,
					lpbihIn->biClrImportant
			   );

	    Dprintf(	TEXT("Output bitmap info header\n\n")
					TEXT("biSize:\t\t%lu\n")
					TEXT("biWidth:\t%ld\n")
					TEXT("biHeight:\t%ld\n")
					TEXT("biPlanes:\t%lu\n")
					TEXT("biBitCount:\t%lu\n")
					TEXT("biCompression:\t%lu (0x%08.8lx) %4.4hs\n")
					TEXT("biSizeImage:\t%lu\n")
					TEXT("biXPelsPerMeter:\t%ld\n")
					TEXT("biYPelsPerMeter:\t%ld\n")
					TEXT("biClrUsed:\t%lu\n")
					TEXT("biClrImportant:\t%lu\n\n"),
					lpbihOut->biSize,
					lpbihOut->biWidth,
					lpbihOut->biHeight,
					lpbihOut->biPlanes,
					lpbihOut->biBitCount,
					lpbihOut->biCompression, lpbihOut->biCompression,
											&lpbihOut->biCompression,
					lpbihOut->biSizeImage,
					lpbihOut->biXPelsPerMeter,
					lpbihOut->biYPelsPerMeter,
					lpbihOut->biClrUsed,
					lpbihOut->biClrImportant
			   );
		return(1L);
	}

	if ((hr = ICClose(hIC)) != ICERR_OK)
	{
		Eprintf(TEXT("Error code %ld closing compressor \"%4.4hs\"\n"), hr,
										&lpbihOut->biCompression);
		return(1L);
	}

	return(0L);
}


//
//  FUNCTION: FindVideoCompressor
//
//  PURPOSE:
//    emulates ICLocate	(just so it's clearer what's going on)
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    Zero on success, non-zero on error
//
//  COMMENTS:
//    Based on WalkStuff() from the ICWalk sample
//

HIC FindVideoCompressor
(
	LPBITMAPINFOHEADER	lpbihIn,
	LPBITMAPINFOHEADER	lpbihOut,
	WORD				wFlags
)
{
	HIC hic;
	int i;
	ICINFO    icinfo;


	// call ICInfo to enumerate compressors.

	Dprintf(TEXT("FindVideoCompressor()\n"));
    Dprintf(	TEXT("Input bitmap info header\n\n")
				TEXT("biSize:\t\t%lu\n")
				TEXT("biWidth:\t%ld\n")
				TEXT("biHeight:\t%ld\n")
				TEXT("biPlanes:\t%lu\n")
				TEXT("biBitCount:\t%lu\n")
				TEXT("biCompression:\t%lu (0x%08.8lx) %4.4hs\n")
				TEXT("biSizeImage:\t%lu\n")
				TEXT("biXPelsPerMeter:\t%ld\n")
				TEXT("biYPelsPerMeter:\t%ld\n")
				TEXT("biClrUsed:\t%lu\n")
				TEXT("biClrImportant:\t%lu\n\n"),
				lpbihIn->biSize,
				lpbihIn->biWidth,
				lpbihIn->biHeight,
				lpbihIn->biPlanes,
				lpbihIn->biBitCount,
				lpbihIn->biCompression, lpbihIn->biCompression,
										&lpbihIn->biCompression,
				lpbihIn->biSizeImage,
				lpbihIn->biXPelsPerMeter,
				lpbihIn->biYPelsPerMeter,
				lpbihIn->biClrUsed,
				lpbihIn->biClrImportant
		   );

    Dprintf(	TEXT("Output bitmap info header\n\n")
				TEXT("biSize:\t\t%lu\n")
				TEXT("biWidth:\t%ld\n")
				TEXT("biHeight:\t%ld\n")
				TEXT("biPlanes:\t%lu\n")
				TEXT("biBitCount:\t%lu\n")
				TEXT("biCompression:\t%lu (0x%08.8lx) %4.4hs\n")
				TEXT("biSizeImage:\t%lu\n")
				TEXT("biXPelsPerMeter:\t%ld\n")
				TEXT("biYPelsPerMeter:\t%ld\n")
				TEXT("biClrUsed:\t%lu\n")
				TEXT("biClrImportant:\t%lu\n\n"),
				lpbihOut->biSize,
				lpbihOut->biWidth,
				lpbihOut->biHeight,
				lpbihOut->biPlanes,
				lpbihOut->biBitCount,
				lpbihOut->biCompression, lpbihOut->biCompression,
										&lpbihOut->biCompression,
				lpbihOut->biSizeImage,
				lpbihOut->biXPelsPerMeter,
				lpbihOut->biYPelsPerMeter,
				lpbihOut->biClrUsed,
				lpbihOut->biClrImportant
		   );

	for (i = 0; ICInfo(ICTYPE_VIDEO, i, &icinfo); i++)
	{
		Dprintf(TEXT("i = %d\n"), i);

		hic = ICOpen(ICTYPE_VIDEO, icinfo.fccHandler, ICMODE_QUERY);

		if (hic)
		{
			Dprintf(TEXT("Got handle\n"));

			if (ICCompressQuery(hic, lpbihIn, lpbihOut) == ICERR_OK)
			{
					Dprintf(TEXT("Query was OK\n"));
				    ICClose(hic);
					hic = ICOpen(ICTYPE_VIDEO, icinfo.fccHandler, wFlags);
					return(hic);
			}
		    ICClose(hic);
		}
	}

	return(NULL);
}
